from pyrasfile.profileWriter import *
from pyrasfile.reportReader import *
from pyrasfile.utils import *
from pyrasfile.csvReader import *
import sys

# Entries to look for
ENTRIES = ["Q Total (cfs)", "Avg. Vel. (ft/s)", "Max Chl Dpth (ft)", "Shear (lb/sq ft)"]

# File path for current development
PATH = "V:\\LosAngelesProjectsData\\HEC-RAS\\Full MODEL\\FullModel-HF.rep"
OUTPATH = "V:\\LosAngelesProjectsData\\HEC-RAS\\Docs_Results\\HFResults.csv"


# Nodes to look for
NODES = [
    riverNode("Upper LA River", "Above RH", "70185*", "F34D"),
    riverNode("Rio Hondo Chnl", "RHC", "7000", "F45B"),
    riverNode("Upper LA River", "Above RH", "157606.4", "GLEN"),
    riverNode("LA River", "Below CC", "21500", "F319"),
    riverNode("Compton Creek", "CC", "22616", "F37B"),
    riverNode("Upper LA River", "Above RH", "195289.1", "F300"),
    riverNode("Upper LA River", "Above RH", "168989.7", "LA14"),
    riverNode("Upper LA River", "Above RH", "199036.7", "LA17"),
    riverNode("LA River", "Below CC", "15700", "LA2"),
    riverNode("Rio Hondo Chnl", "RHC", "15600", "RHD2"),
    riverNode("Upper LA River", "Above RH", "213974.*", "LA19"),
    riverNode("Upper LA River", "Above RH", "207769.0", "LA18"),
    riverNode("Upper LA River", "Above RH", "186920.5", "LA16"),
    riverNode("Upper LA River", "Above RH", "177244.0", "LA15"),
    riverNode("Upper LA River", "Above RH", "161809.0", "LA13"),
    riverNode("Upper LA River", "Above RH", "150089.6", "LA12"),
    riverNode("Upper LA River", "Above RH", "143272.1", "LA11"),
    riverNode("Upper LA River", "Above RH", "137061.4", "LA10"),
    riverNode("Upper LA River", "Above RH", "115507", "LA8"),
    riverNode("Upper LA River", "Above RH", "107352", "LA7"),
    riverNode("Upper LA River", "Above RH", "88311", "LA5"),
    riverNode("Upper LA River", "RH to CC", "52928", "LA4"),
    riverNode("Upper LA River", "RH to CC", "38700", "LA3"),
    riverNode("Upper LA River", "Above RH", "96076", "LA6"),
    riverNode("LA River", "Below CC", "1600", "LA1"),
    riverNode("Compton Creek", "CC", "33669.4*", "CP2"),
    riverNode("Compton Creek", "CC", "44173.9*", "CP3"),
    riverNode("Upper LA River", "Above RH", "127537.6", "LA9"),
    riverNode("Compton Creek", "CC", "52494.08", "CP4"),
    riverNode("Rio Hondo Chnl", "RHC", "30500", "RHD3"),
    riverNode("Upper LA River", "Above RH", "225805.0", "11092450"),
    riverNode("Rio Hondo Chnl", "RHC", "44113", "11102300"),
    riverNode("Upper LA River", "Above RH", "128608", "F57C"),
    riverNode("Compton Creek", "CC", "7927", "CP1"),
    riverNode("Rio Hondo Chnl", "RHC", "914", "RHD1")
]

flowNodes = [
    riverNode("Compton Creek", "CC", "52494.08"),
    riverNode("LA River", "Below CC", "29266"),
    riverNode("Rio Hondo Chnl", "RHC", "44113"),
    riverNode("Upper LA River", "Above RH", "225805.0"),
    riverNode("Upper LA River", "RH to CC", "63900.3*")
]


flowRanges = [1, 10, 100, 1000, 10000]
upstreamFlows = [[], [], []]
for i in flowRanges:
    for j in flowRanges:
        for k in flowRanges:
            upstreamFlows[0].append(i)
            upstreamFlows[1].append(j)
            upstreamFlows[2].append(k)

downstreamFlows = [
    [upstreamFlows[0][i] + upstreamFlows[1][i] for i in range(0, 125)],
    [upstreamFlows[0][i] + upstreamFlows[1][i] + upstreamFlows[2][i] for i in range(0, 125)]
]

"""
The flow range at the bottom appears to be about up to 70,000 cfs; let's say up to
50,000 cfs in each of the upstream reaches.  If I run 100 profiles on a log scale,
this makes powers of about 1.12.
"""

upstreamFlowRange = [1.1 ** i for i in range(0, 100)]

flowdata = {}
for node in flowNodes:
    river = node["river"]
    reach = node["reach"]
    rs = node["rs"]
    key = mkFlowHeader(river, reach, rs)
    # Upstream and downstream flows
    flows = {
        "Above RH": upstreamFlowRange,
        "RHC": upstreamFlowRange,
        "CC": upstreamFlowRange,
        "RH to CC": [2 * i for i in upstreamFlowRange],
        "Below CC": [3 * i for i in upstreamFlowRange]
    }[reach]
    flowdata[key] = flows

"""flowdata = {
    "River Rch & RM=Compton Creek,CC              ,52494.08": upstreamFlows[0],
    "River Rch & RM=LA River,Below CC        ,29266": downstreamFlows[1],
    "River Rch & RM=Rio Hondo Chnl,RHC             ,44113": upstreamFlows[1],
    "River Rch & RM=Upper LA River,Above RH        ,225805.0": upstreamFlows[2],
    "River Rch & RM=Upper LA River,RH to CC        ,63900.3*": downstreamFlows[0]
}"""

def mkRegBound(pn, flow):
    return mkBoundaryData("Junction", "Junction", "", "")

def mkBottomBound(pn, flow):
    return mkBoundaryData("Junction", "Normal Depth", "", "0.001")
    # F319 seems to always be around 1.3-1.5 ft deep
    # F319 = LAR 21500, just a few miles above the estuary, so probably a reasonable basis

bounds = {
    "Compton Creek,CC": mkRegBound,
    "LA River,Below CC": mkBottomBound,
    "Rio Hondo Chnl,RHC": mkRegBound,
    "Upper LA River,Above RH": mkRegBound,
    "Upper LA River,RH to CC": mkRegBound
}

text = buildFile(100, flowdata, bounds, title="GenFlow Preliminary 1-100")

if __name__ == "__main__":
    which = ""
    if len(sys.argv) > 1:
        which = sys.argv[1]
    else:
        print("Generate: -g; parse: -p; make CSV: -m; generate from CSV: -gr")
    generate = which == "-g" or which == "-gr"
    parse = which == "-p"
    makeCSV = which == "-m"
    readCSV = which == "-gr"
    if generate:
        if readCSV:
            text = buildFile(88, csvToFlowData("Z:\\adit\\Desktop\\LARFlows\\code\pyRasFile\\empiricalFlows.csv"), bounds, title="Empirical Flows 2-88")
            # text = buildFile(100, flowdata, bounds, title = "Flow Range 10k - 100")
        with open("V:\\LosAngelesProjectsData\\HEC-RAS\\Full Model\\FullModel.f05", "w") as f:
            f.write(text)
    if parse:
        convertCSV(NODES, entries = ENTRIES, inpath = PATH, outpath = OUTPATH, selective = True, swmm = True)
    if makeCSV:
        flows = [1, 10, 100, 1000, 10000]
        nodes = [
            {"river": "Compton Creek", "reach": "CC", "rs": "52494.08"},
            {"river": "LA River", "reach": "Below CC", "rs": "29266"},
            {"river": "Rio Hondo Chnl", "reach": "RHC", "rs": "44113"},
            {"river": "Upper LA River", "reach": "Above RH", "rs": "225805.0"},
            {"river": "Upper LA River", "reach": "RH to CC", "rs": "63900.3*"}
        ]
        upstreamNodes = {
            mkFlowHeader(nodes[1]["river"], nodes[1]["reach"], nodes[1]["rs"]):
                [mkFlowHeader(nodes[0]["river"], nodes[0]["reach"], nodes[0]["rs"]),
                                     mkFlowHeader(nodes[2]["river"], nodes[2]["reach"], nodes[2]["rs"]),
                                     mkFlowHeader(nodes[3]["river"], nodes[3]["reach"], nodes[3]["rs"])],
            mkFlowHeader(nodes[4]["river"], nodes[4]["reach"], nodes[4]["rs"]):
                [mkFlowHeader(nodes[2]["river"], nodes[2]["reach"], nodes[2]["rs"]),
                                     mkFlowHeader(nodes[3]["river"], nodes[3]["reach"], nodes[3]["rs"])]
        }
        path = "Z:\\adit\\Desktop\\LARFlows\\code\\pyRasFile\\FlowPerms.csv"
        perms = generatePermutedFlows(flows, nodes, upstreamNodes, write = True, path = path, debug = True)
        for key in perms:
            print("%s: %d" % (key, len(perms[key])))



"""
Written beginning July 18, 2019 by Daniel Philippus for the LAR Environmental Flows project at Colorado School of Mines.

This component of the program writes a set of flows to a HEC-RAS steady state flow file under the relevant profiles.

Between this and reportReader, a set of profiles can be automatically generated, after which HEC-RAS is run, a report
generated, and the report automatically converted into a machine-friendly CSV format for further processing (e.g.
statistical analysis with R).
"""

"""
HEC-RAS flow file format (saved as *.f01, etc):

Flow Title=Flow 01
Program Version=5.07
Number of Profiles= 125
Profile Names=PF 1,PF 2,PF 3,...
River Rch & RM=Compton Creek,CC              ,52494.08
       1       1       1       1       1       1       1       1       1       1
       1       1       1       1       1       1       1       1       1       1
       1       1       1       1       1      10      10      10      10      10
      10      10      10      10      10      10      10      10      10      10
      10      10      10      10      10      10      10      10      10      10
     100     100     100     100     100     100     100     100     100     100
     100     100     100     100     100     100     100     100     100     100
     100     100     100     100     100    1000    1000    1000    1000    1000
    1000    1000    1000    1000    1000    1000    1000    1000    1000    1000
    1000    1000    1000    1000    1000    1000    1000    1000    1000    1000
   10000   10000   10000   10000   10000   10000   10000   10000   10000   10000
   10000   10000   10000   10000   10000   10000   10000   10000   10000   10000
   10000   10000   10000   10000   10000
<Other reaches>
Boundary for River Rch & Prof#=Compton Creek,CC              , 1 
Up Type= 0 
Dn Type= 0
<Other reaches>
DSS Import StartDate=
DSS Import StartTime=
DSS Import EndDate=
DSS Import EndTime=
DSS Import GetInterval= 0 
DSS Import Interval=
DSS Import GetPeak= 0 
DSS Import FillOption= 0 
"""

"""
Format notes:
* The end of one number is always exactly 8 characters to the right of the end of the previous (columns width 8), i.e.
    the last digit is always the 8th character
* There are at most 10 columns
* The comma before the profile number in the boundary condition is 17 columns to the right of the comma
    before the reach name, e.g. in columns 45 and 62 respectively, i.e. 16 characters (including reach name) in between
* The boundary condition type 0 means junction or no particular boundary condition (for upstream reaches)
* BC 1,2,3,4 presumably mean, respectively, Known W.S., Critical Depth, Normal Depth, and Rating Curve.  Formats
are as follows:

Known WS:
Up Type= 0 
Dn Type= 1 
Dn Known WS=10

Crit Depth:
Up Type= 0 
Dn Type= 2 

Norm. Depth:
Up Type= 0 
Dn Type= 3 
Dn Slope=0.05

Rating Curve doesn't seem to be specified in the flow profiles file.

Note that the boundary conditions are duplicated for each profile, thus a long sequence of:

Boundary for River Rch & Prof#=Upper LA River,RH to CC        , 125 
Up Type= 0 
Dn Type= 0 
"""

# I think this is just for unsteady flow, no need for now to support modifications to it
FILE_END = """DSS Import StartDate=
DSS Import StartTime=
DSS Import EndDate=
DSS Import EndTime=
DSS Import GetInterval= 0 
DSS Import Interval=
DSS Import GetPeak= 0 
DSS Import FillOption= 0"""

def mkBoundaryHeader(river, reach, profile):
    # River/reach/profile specification for boundary conditions
    return "Boundary for River Rch & Prof#=%s,%s%s, %d" % (river,reach, " " * (16 - len(reach)), profile)

def mkBoundaryData(upname, dname, uparam="", dparam=""):
    # boundary condition data
    types = {"Known WS": 1, "Critical Depth": 2, "Normal Depth": 3, "Junction": 0}
    data = {1: "Known WS", 3: "Slope"}
    utype = types[upname]
    dtype = types[dname]
    items = ["Up Type= %d" % utype, "Dn Type= %d" % dtype]
    if utype in [1, 3]:
        items.append("Up %s=%s" % (data[utype], uparam))
    if dtype in [1,3]:
        items.append("Dn %s=%s" % (data[dtype], dparam))
    return items

def mkBoundary(river, reach, profile, upname, dname, uparam, dparam):
    # Full boundary condition
    return "\n".join([mkBoundaryHeader(river, reach, profile)] + mkBoundaryData(upname, dname, uparam, dparam))

def mkFlowHeader(river, reach, station):
    # River, reach, station specification for flow data
    return "River Rch & RM=%s,%s%s,%s" % (river, reach, " " * (16 - len(reach)), station)

def mkFlowData(flows):
    # Flow data
    flowStrs = ["%5.1f" % flow for flow in flows] # Avoid the strings being too long
    flowStrs = [" " * (8-len(flow)) + flow for flow in flowStrs]
    flowRows = []
    prev = 0
    for ix in range(1, len(flowStrs)):
        if ix % 10 == 0:
            flowRows.append(flowStrs[prev:ix])
            prev = ix
    if prev < len(flowStrs) - 1:
        flowRows.append(flowStrs[prev:])
    return ["".join(row) for row in flowRows]

def mkHeader(nprofiles, title="Flow 01", ver="5.0.7"):
    # File header - title,version,profile information
    return "Flow Title=%s\nProgram Version=%s\nNumber of Profiles= %d\nProfile Names=%s" % (
        title,
        ver,
        nprofiles,
        ",".join(["PF %d" % pn for pn in range(1, nprofiles + 1)])
    )

def buildFile(nprofiles, profiledata, bounddata, title="Flow 01", ver="5.0.7", end=FILE_END):
    """
    profiledata format: a dictionary of the necessary data.
        {flowheader: [flows]}
    bounddata: a dictionary of the necessary boundary condition generators, as functions
        {"River,Reach": function}
            The function accepts the profile number and the flow volume (though it need not use the latter),
            and returns the appropriate boundary data as a string
    """
    header = mkHeader(nprofiles, title, ver)
    flowdata = []
    bounds = []
    # Build flow data and boundary data
    for pheader in profiledata.keys():
        flowdata.append(pheader)
        flows = profiledata[pheader]
        if len(flows) != nprofiles:
            raise ValueError("Number of flow profiles given does not match specified profile count!")
        fd = mkFlowData(flows)
        flowdata = flowdata + fd
        boundspec = pheader.split("=")[1].split(",")[0:2] # This grabs the river and reach
        boundspec[1] = boundspec[1].strip()
        for pn in range(0, nprofiles):
            bheader = mkBoundaryHeader(boundspec[0], boundspec[1], pn + 1)
            bdata = bounddata[",".join(boundspec)](pn + 1, flows[pn])
            bounds.append(bheader)
            bounds = bounds + bdata
    return "\n".join([header] + flowdata + bounds + [end])




"""
Written beginning July 22, 2019 by Daniel Philippus for the LAR Environmental Flows project at Colorado School of Mines.

This file is a collection of simple utilities to aid in tasks encountered with this semi-automated HEC-RAS usage
that are not large or complex enough to merit their own module.
"""

from pyrasfile.profileWriter import mkFlowHeader
from itertools import permutations

def generatePermutedFlows(flows, nodes, upstreamNodes, write = False, path = "", debug = False):
    """
    Generate all possible permutations of the given flows across the relevant nodes,
    with the constraint that downstream nodes will have a flow that is the sum of the nodes
    upstream of them.  Optionally, write to a CSV compatible with the format specified in
    csvReader.py for later use.
    
    :param flows: A list of the flow rates (in cfs) to use
    :param nodes: A list of the nodes, each node being a dictionary of "river", "reach", and "rs"
    :param upstreamNodes: A dictionary of which nodes are upstream of which, in the format {node: [upstream nodes]},
                            where the nodes are in the format given by mkFlowHeader from profileWriter.py.
                            Note that the upstream nodes must be a list of the furthest-upstream nodes, none of which
                            are themselves downstream of any other nodes, or it will not work.
    :param write: Whether to write to a CSV file.
    :param path: The CSV path; required if write = True.
    :return: The flow data in the standard flow-data format used by the package, which is {node: [flows]}
    """

    nodeKeys = {}
    for node in nodes:
        key = mkFlowHeader(node["river"], node["reach"], node["rs"])
        nodeKeys[key] = {}
        # Need to keep the original data around for writing as CSV
        nodeKeys[key]["data"] = node
    # Nodes that don't have any nodes upstream of them
    noUpstream = [node for node in nodeKeys.keys() if (node not in upstreamNodes.keys()) or upstreamNodes[node] == []]
    # Number of unconstrained nodes
    nuc = len(noUpstream)
    nflows = len(flows)
    nperms = nflows ** nuc
    if debug:
        print("N. Permutations: %d" % nperms)
    # perms = permutations(flows, nuc)
    perms = [[]] * nuc
    """
    The idea here:
        First node: each flow in order repeated nperms / n. flows times, run once
        Second node: each flow in order repeated nperms / (n. flows)^2 times, run n. flows times
        nth node: each flow in order repeated nperms / (n. flows)^n times, run n.flows ^ n times
    Therefore, we iterate through the number of permutations first (0 to n.flows - 1), and within each loop,
    we iterate through the number of flows.
    
    For each entry in the permutations, the index in the flows needs to be (0 * nreps), (1 * nreps), etc, which should
    be... Let's look at it with 2 flows and 3 upstream locations.  Our permutations of indices are:
    0 0  0 0  1 1  1 1
    0 0  1 1  0 0  1 1
    0 1  0 1  0 1  0 1
    
    So for the first row, it's index // 4 = index // (nflows ^ 2) = index // (nflows ^ (nuc - 1))
    Second row, it's index // 2 mod 2 = index // nflows mod nflows = index // (nflows ^ (nuc - 2)) mod nflows
    Third row, it's index mod 2 = index // (nflows ^ (nuc - 3)) mod nflows
    
    So our overall formula should be... (index // (nflows ^ (nuc - row - 1))) mod nflows.
    
    This runs across the number of rows nuc, and within each row, index iterates from 0 to nperms
    """
    # Oddly, each index of permutations gets longer -- e.g. 125, 250, 375.  Why?
    # It seems like each set of these may be getting appended to *all* sets of
    # permutations.  This makes no sense at all.  It is what is happening, however,
    # as all the permutations end up the same.
    for nn in range(0, nuc):
        for ix in range(0, nperms):
            index = (ix // (nflows ** (nuc - nn - 1))) % nflows
            # Yes, this is weird, but otherwise it was appending it to *everything*
            # Because Pythonic weirdness, I guess?  I've seen it behave oddly before,
            # just not in this particular way.
            perms[nn] = perms[nn] + [flows[index]]
    if debug:
        for perm in perms:
            print("Permutation length: %d" % len(perm))
        permsEqual = True
        for perm in perms:
            for perm2 in perms:
                if perm != perm2:
                    permsEqual = False
        print("Permutations are equal: %s" % str(permsEqual))

    # Assign each set of flows to a node
    # BUG: for some reason these are getting written to 3x more than they should, and it's happening
    # here at the latest
    index = 0
    for key in noUpstream:
        nodeKeys[key]["flows"] = perms[index]
        index += 1

    # Now, set up the downstream nodes
    for key in nodeKeys.keys():
        if key in upstreamNodes.keys():
            nodeKeys[key]["flows"] = []
            upstream = upstreamNodes[key]
            for ix in range(0, nperms):
                # Sum of the upstream flows
                nodeKeys[key]["flows"].append(sum([nodeKeys[u]["flows"][ix] for u in upstream]))
    # Need to restructure the dict to be {node key: [flows]}
    output = {}
    # for key in nodeKeys.keys():
    for key in noUpstream:
        output[key] = nodeKeys[key]["flows"]

    # Set up the CSV, if needed
    if write:
        with open(path, "w") as outfile:
            lines = ["river,reach,rs,profilenumber,flow"]
            for entry in nodeKeys.values():
                data = entry["data"]
                datastr = "%s,%s,%s" % (data["river"], data["reach"], data["rs"])
                for index, flow in enumerate(entry["flows"]):
                    # Index is for the profile number
                    lines.append("%s,%d,%f" % (datastr, index + 1, flow))
            outfile.write("\n".join(lines))
    return output


"""
Written beginning July 22, 2019 by Daniel Philippus for the LAR Environmental Flows project at Colorado School of Mines.

This component of the program parses a CSV file to produce the appropriate flow inputs for the profile writer.  As
of the present state of development, the user must still manually specify boundary conditions and the other inputs.
The flow data is expected to typically be the most complex, and therefore the part which would benefit the most from
automated input from a format which other programs can conveniently generate.

The CSV format is expected to be as follows:

river,reach,rs,profilenumber,flow

Where flow is a floating-point value in cubic feet per second, profilenumber is an integer,
and the remaining entries are strings.

The order of these can be varied (if specified in the relevant functions or if a header is included),
but the names must remain the same.
"""

from pyrasfile.profileWriter import mkFlowHeader

def readCSVList(path):
    # Utility function - convert CSV into list
    with open(path, "r") as f:
        # stripping in case of Windows-style (cr-lf) newlines
        return [l.strip("\r").split(",") for l in f.read().split("\n")]

def parseCSV(csvList, header = True, columns = [], types = False, coltypes = {}, allowEmpty = False, emptyCols = []):
    """
    Given a CSV parsed into a list of lists, convert it into a list of dictionaries, either with the first line
    as headers or with a specified column order.

    If types is true, it will also parse the entries according to the parsing functions in coltypes; if not,
    it will leave them all as strings.  coltypes must be specified if types is True.  coltypes should be a dictionary
    of {column: parser}, e.g. {"profilenumber": int}.

    The empty-related parameters are ignored unless types is True.  If types are being used, then, if an entry is empty:
        * If allowEmpty is false, then the function will throw an error if the column name is not in emptyCols, or
            if it is, then the value will be set to None.
        * If allowEmpty is true, then the value will be set to None.
    This does not apply if the specified type is string.
    """
    if header:
        columns = csvList[0]
    cols = len(columns)
    result = []
    start = 1 if header else 0
    for row in csvList[start:]:
        count = len(row)
        if count != cols:
            if row == [] or row == [""]:
                # Empty row, just skip it
                continue
            raise ValueError("Error: number of columns is not consistent in row: %s" % row)
        out = {}
        for i in range(0, count):
            val = row[i]
            col = columns[i]
            if types:
                if val == "":
                    if coltypes[col] == str:
                        out[col] = val
                    elif allowEmpty == True or col in emptyCols:
                        out[col] = None
                    else:
                        raise ValueError("Error: empty entry not allowed in row: %s" % row)
                else:
                    out[col] = coltypes[col](val)
            else:
                out[col] = val
        result.append(out)
    return result

def parseCSVfile(path, toDict=True, header=True, columns=[], types=False, coltypes={}, allowEmpty=False, emptyCols=[]):
    # Parse from a file and read into a dictionary by default; if toDict = False, this just calls
    # readCSVList
    if toDict:
        return parseCSV(readCSVList(path), header, columns, types, coltypes, allowEmpty, emptyCols)
    else:
        return readCSVList(path)

def parseFlowCSV(path, header = True, columns = ["river", "reach", "rs", "profilenumber", "flow"]):
    coltypes = {
        "river": str,
        "reach": str,
        "rs": str,
        "profilenumber": int,
        "flow": float
    }
    return parseCSVfile(path, header = header, columns = columns, types = True, coltypes = coltypes)

def makeFlowData(csvDict):
    result = {}
    for entry in csvDict:
        key = mkFlowHeader(entry["river"], entry["reach"], entry["rs"])
        if not key in result.keys():
            # the final format will be a list of lists, but we need to keep the profile numbers for now
            # to make sure the order is correct
            result[key] = {entry["profilenumber"]:entry["flow"]}
        else:
            result[key][entry["profilenumber"]] = entry["flow"]
    for key in result.keys():
        # Make sure they're in the correct order
        entry = result[key]
        keys = list(entry.keys())
        keys.sort()
        result[key] = [entry[k] for k in keys]
    return result

def csvToFlowData(path, header = True, columns = ["river", "reach", "rs", "profilenumber", "flow"]):
    return makeFlowData(parseFlowCSV(path, header, columns))



"""
Written beginning July 17, 2019 by Daniel Philippus for the LAR Environmental Flows project at Colorado School of Mines.

This component of the program reads HEC-RAS Report (.rep) files to extract relevant information and can, optionally,
write the data to a CSV file (e.g. for use in R scripts).

It is designed to look specifically for cross sections by river and reach, but within cross-section data is designed
to be extensible in terms of which data it looks for.  While the data is simply specified in a variable as only specific
data is necessary as of the program's writing, this should be easy to modify if necessary.
"""

"""
HEC-RAS Report Format for relevant entries (example taken from a .rep file):
This is repeated for each cross-section.

CROSS SECTION          


RIVER: Compton Creek   
REACH: CC                 RS: 52494.08


CROSS SECTION OUTPUT  Profile #PF 1  
                                                                                               
  E.G. Elev (ft)            114.48    Element                   Left OB    Channel   Right OB  
  Vel Head (ft)               0.35    Wt. n-Val.                            0.011              
  W.S. Elev (ft)            114.13    Reach Len. (ft)            29.51      29.51      29.51   
  Crit W.S. (ft)            114.13    Flow Area (sq ft)                     20.99              
  E.G. Slope (ft/ft)      0.002285    Area (sq ft)                          20.99              
  Q Total (cfs)             100.00    Flow (cfs)                           100.00              
  Top Width (ft)             30.00    Top Width (ft)                        30.00              
  Vel Total (ft/s)            4.76    Avg. Vel. (ft/s)                       4.76              
  Max Chl Dpth (ft)           0.70    Hydr. Depth (ft)                       0.70              
  Conv. Total (cfs)         2092.2    Conv. (cfs)                          2092.2              
  Length Wtd. (ft)           29.51    Wetted Per. (ft)                      31.40              
  Min Ch El (ft)            113.43    Shear (lb/sq ft)                       0.10              
  Alpha                       1.00    Stream Power (lb/ft s)                 0.45              
  Frctn Loss (ft)             0.07    Cum Volume (acre-ft)                  44.31              
  C & E Loss (ft)             0.00    Cum SA (acres)                        58.14              
                                                                                               

Warning: The energy equation could not be balanced within the specified number of iterations.  The program used critical 
         depth for the water surface and continued on with the calculations.
Warning: During the standard step iterations, when the assumed water surface was set equal to critical depth, the calculated 
         water surface came back below critical depth.  This indicates that there is not a valid subcritical answer.  The 
         program defaulted to critical depth.

CROSS SECTION OUTPUT  Profile #PF 2  
                                                                                               
<... same items as PF 1>
CROSS SECTION OUTPUT  Profile #PF 3  
                                                                                               
<... same items as PF 1>
Warning: The cross-section end points had to be extended vertically for the computed water surface.
"""

def riverNode(river, reach, rs, swmm = ""):
    # This just makes a dict; defining a function purely for convenience
    # swmm = corresponding SWMM node; this is specific to the LARFlows project and can be ignored by other users
    return {"river": river, "reach": reach, "rs": rs, "swmm": swmm}



def getDataForNodes(xsData, nodes):
    # Extract the relevant data from the parsed-out data
    keys = [" ".join([node["river"], node["reach"], node["rs"]]) for node in nodes]
    availKeys = [key for key in keys if key in xsData.keys()]
    entries = [xsData[key] for key in availKeys]
    for key in [key for key in keys if not key in availKeys]:
        print("Warning: node %s not found in report data--skipping." % key)
    return entries

def makeNodePfDataString(nodeData, pf, riverNode, entries, swmm = False, nodes = []):
    # Make a CSV line of the relevant data from the node, for the given profile number
    # In order to not be selective, just set entries to be all of the keys for an arbitrary cross-section;
    # specifying entries is necessary, however, to ensure a consistent order
    # If swmm, also include the SWMM node
    outData = [riverNode["river"], riverNode["reach"], riverNode["rs"], pf]
    for entry in entries:
        if entry in nodeData.keys():
            outData.append(nodeData[entry])
        else:
            outData.append("")
    # Find what SWMM node it corresponds to, if necessary
    if swmm:
        found = False
        for node in nodes:
            if found:
                break
            if node["river"] == riverNode["river"] and node["reach"] == riverNode["reach"] and node["rs"] == riverNode["rs"]:
                found = True
                outData.append(node["swmm"])
        if not found:
            outData.append("")
    return ",".join(outData)

def buildCSV(xsData, nodes, entries, selective = False, swmm = False):
    # Build the CSV file for the relevant nodes and, if selective, relevant entries
    # If not selective, entries will simply be the entries of the first node
    # If swmm, data will also include which swmm node the entry corresponds to
    entriesSet = selective
    data = getDataForNodes(xsData, nodes)
    output = []
    for datum in data:
        pfs = [k for k in datum.keys() if not k in ["river", "reach", "rs"]] # PF numbers only
        for pf in pfs:
            nodeData = datum[pf]
            if not entriesSet:
                entries = nodeData.keys() # To keep a specific order
                entriesSet = True
            output.append(makeNodePfDataString(nodeData, pf, datum, entries, swmm, nodes))
    output = ["River,Reach,RS,Profile," + ",".join(entries) + ",SWMM Node" if swmm else ""] + output
    return "\n".join(output)

def convertCSV(nodes, entries, inpath, outpath, selective = False, swmm = False):
    with open(outpath, "w") as f:
        f.write(buildCSV(parseFile(getReportFile(inpath)), nodes, selective = selective, entries = entries, swmm = swmm))

def getReportFile(filename):
    with open(filename, "r") as f:
        return f.read()

def crossSections(text):
    # Split file content into separate cross sections
    # CROSS SECTION on its own shows up a lot, but it's only followed by two blank spaces when it's a new XS
    return text.split("CROSS SECTION  ")

def nodeData(xs):
    # Get the node information (river, reach, rs) for a given cross-section
    # Information is on the first two lines after CROSS SECTION
    lines = [ln for ln in xs.split("\n") if ln.strip() != ""][0:2]
    river = " ".join([i for i in lines[0].split(" ") if i != ""][1:])   # The first thing is RIVER:;
                                                                        # the rest is the river name
    ln2dat = [i for i in lines[1].split(" ") if i != ""]
    ln2splitIndex = ln2dat.index("RS:") # Where the reach stops and the RS begins
    reach = " ".join(ln2dat[1:ln2splitIndex]) # Everything before RS: except for REACH:, which starts the line
    rs = ln2dat[ln2splitIndex + 1] # No spaces in RS, it's a number
    return {"river": river, "reach": reach, "rs": rs}

def profiles(xs):
    # Split a cross section into profiles
    return xs.split("Profile #PF")

def entries(profile):
    # Data starts on row 3 and continues through row 17
    data = profile.split("\n")[2:17]
    output = {}
    for row in data:
        items = [i.strip() for i in row.split("  ") if i != ""] # Separate items are always separated by multiple spaces
        # There are two rows that have three entries for the right-hand items, but we don't need either, so
        # we can ignore that.
        # Otherwise, they are all two key-value pairs.
        if len(items) >= 4: # Sometimes something seems to be missing from the row
            output[items[0]] = items[1]
            output[items[2]] = items[3]
    return output

def parseXs(xs):
    # Parse the cross-section, returning node data, profile number, and entries
    data = nodeData(xs)
    profs = profiles(xs)[1:]
    for prof in profs:
        pnum = prof.split("\n")[0].strip() # First line has the profile number
        data[pnum] = entries(prof)
    return data

def parseFile(text):
    # Extract cross-section data from file
    xses = crossSections(text)
    xses = [xs for xs in xses if "CROSS SECTION OUTPUT" in xs] # Filter out non-cross-section entries at beginning
    data = {}
    while len(xses) > 0:
        xs = xses[0]
        xsData = parseXs(xs)
        data[" ".join([xsData["river"], xsData["reach"], xsData["rs"]])] = xsData
        # Shrink the list as the program goes through it, in order to reduce memory usage
        xses = xses[1:]
    return data



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()


setuptools.setup(
    name="pyrasfile",
    version="1.0.0",
    author="Daniel Philippus",
    author_email="dphilippus@protonmail.com",
    description="A collection of HEC-RAS file writers and parsers",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/larflows/pyrasfile",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: Microsoft :: Windows",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
    ]
)

